/*  mdate : Fetch time from remote host and maybe set ours. 
 *
 *
 * Copyright (c) 1995 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.

 **********************************************************************
                               AUTHORS
 **********************************************************************

 * This program contains parts of   PING.C   written by:
 * Mike Muuss,  U. S. Army Ballistic Research Laboratory
 * December, 1983  Modified at Uc Berkeley
 *
 * and parts of   RDATE.C :
 *
 *	(c) Copyright 1989 by the Trustees of Dartmouth College.
 *	This program may be copied and used for any non-commercial
 *	purpose so long as this notice remains attached.
 *
 * Part by R.M.Damerell, Maths Dept, RHBNC, Egham Surrey, UK.
 *
 * Bug reports to RMD (but no promise to fix) at:
 * uhah208@ma.rhbnc.ac.uk
 *
 * Any person accepting or using this program does so entirely at own risk. 
 * No author or distributor accepts any liability for any defects or for
 * any consequential loss or damage however caused.
 *
 * Usage: mdate [options] host ...
 *
 * Mdate sends an  ICMP  timestamp request to the named host(s) and
 * looks for a reply and estimates the time difference.
 * 
 * If the -a or -s flags are set and the caller is the superuser,
 * and if the difference between remote time and local time is
 * more than a specified tolerance (default is 3 msec), then the
 * local host's time is adjusted (-a) or set (-s) to match.
 * 
 * This program should not be expected to be accurate to less
 * than a few milliseconds (in favourable circumstances).
 *
 **********************************************************************
*/

				
				/* INCLUDES AND DEFINES */

#include <stdio.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/param.h>
#include <values.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/signal.h>

#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netdb.h>

#include "site.h"

#ifndef	INADDR_NONE
#define	INADDR_NONE	0xffffffff	/* should be in <netinet/in.h> */
#endif

#define	MAXPACKET	4096	/* max packet size */

#ifndef	MAXHOSTNAMELEN
#define MAXHOSTNAMELEN	  64	/* should be defined in <sys/param.h> */
#endif

#define  K 1000
#define  M 1000000
#define  HUGE  8000000
/* time-differences > HUGE (msec) are almost certainly due to bugs */
#define  SHORT_MASK 0xFFFF	/* 16-bits, all 1 */


#define	SIZE_ICMP_HDR	   8	/* 8-byte ICMP header */
#define	DEF_DATALEN	  56	/* default data area after ICMP header */
#define PACKSIZE          SIZE_ICMP_HDR + DEF_DATALEN

#ifdef OLDSELECT
#define FD_ZERO(sp)	((sp)->fds_bits[0] = 0)
#define FD_SET(fd,sp)	((sp)->fds_bits[0] |= (1<<(fd)))
#endif

/* Smallest time interval we can measure (msec) Must be integer > 0. */
#ifndef QUANTUM
#define QUANTUM 1
#endif

#define DEFTOL 3
#define READTOUT 2

/* Convert struct timeval to milliseconds since midnight */
#define millis(x) ( (long) ( (x.tv_sec  - midnight) * 1000 + x.tv_usec / 1000))

#ifndef max
#define min(a,b) (((a) < (b)) ? (a) : (b))
#define max(a,b) (((a) > (b)) ? (a) : (b))
#endif


				/* PROTOTYPES */

#ifdef __STDC__
# define _ARGS_(s) s
#else
# define _ARGS_(s) ()
#endif

int main _ARGS_((int argc , char **argv )) ;
void doit _ARGS_((int argc , char **argv )) ;
void try_host _ARGS_((void )) ;
int send_packet _ARGS_((int seq )) ;
int in_cksum _ARGS_((u_short *ptr , int nbytes )) ;
int rcv_packet _ARGS_((int seq )) ;
struct icmp *pr_pack _ARGS_((char *buf , int cc , int seq , struct sockaddr_in *from )) ;
char *pr_type _ARGS_((int t )) ;
void set_time _ARGS_((void )) ;

#ifdef NORANDOM
void srandom _ARGS_((int seed )) ;
int random _ARGS_((void )) ;
#endif

char *inet_ntoa _ARGS_(( struct in_addr in ));
unsigned long inet_addr _ARGS_(( char *cp )) ;

#ifndef ntohl
  u_long ntohl _ARGS_(( u_long ));		/* declaration unneeded if ntohl is a macro */
  u_long htonl _ARGS_(( u_long ));
#endif

#undef _ARGS_



				/* GLOBALS */


extern char *optarg;
extern int optind;
extern int opterr ;

int sockfd;			/* socket file descriptor */
char *hostname;
long tsum;			/* sum of all round-trip times, for average */

int dflag = 0, sflag=0, aflag = 0, delay = 0, count=1,
min_trip = MAXINT, best_diff, sum_trip=0, replies=0;

struct sockaddr_in server;	/* who to ping */
int tolerance = DEFTOL;
long midnight;			/* Seconds from epoch to midnight */
n_short ident ;			/* Identify this process */
struct protoent *proto;		/* ICMP protocol */


/*  MAIN. Read command and setup.
 */    
int
main(argc,argv)
     int argc;
     char **argv;
{
  int errflag = 0, c;
  int naptime;
  struct timeval tv ;
  struct tm *tm_p ;

  if ( gettimeofday( &tv, (struct timezone*) 0 ) <0)
    { perror( "mdate: gettimeofday_1"); return(0);}
  tm_p = gmtime((time_t *) &tv.tv_sec) ;
  midnight = tv.tv_sec - 3600*tm_p->tm_hour - 60*tm_p->tm_min - tm_p->tm_sec ;
  if( dflag > 2) printf( "mdate: midnight= %d = %s \n", midnight, ctime(&midnight));

  if ( (proto = getprotobyname("icmp")) == NULL)
    { perror("mdate: getprotobyname: icmp\n"); exit(1);}
  ident = ( n_short) getpid() ;

  opterr  = 0;
  while ((c = getopt(argc,argv,"ac:dt:sw:")) != EOF) 
    {
      switch (c) 
	{
	case 'a':
#ifdef NOADJTIME
	  fprintf(stderr, "mdate: -a option not available.\n");
#else
	  aflag++;
#endif
	  break;
	case 'c':
	  count = atoi(optarg);
	  if (count < 0)
	    errflag++;
	  break;
	case 'd':
	  dflag++;
	  break;
	case 't':
	  tolerance = atoi(optarg);
	  if (tolerance < 0)
	    errflag++;
	  break;
	case 's':
	  sflag++;
	  break;
	case 'w':
	  delay = atoi(optarg);
	  if (delay < 0)
	    errflag++;
	  break;
	default:
	  errflag++;
	}
    }
  if (errflag || optind == argc) 
    {
      fprintf(stderr, "Usage: mdate [-a | -s] [-c int] [-d] [-t int] [-w int] host ...\n" );
      exit(1);
    }

  if (delay )			/* Wait a pseudo random time to avoid collisions */
    {
      srandom( ident );
      naptime = random() & SHORT_MASK ; /* On our Suns, the first random() isnt! */
      naptime = random() & SHORT_MASK ;
      naptime = (int) (  (float) naptime * (float) delay / (float) SHORT_MASK ) ;
      if ( dflag> 1 ) printf( "mdate: waiting %d secs\n", naptime );
      sleep( naptime);
    }

  bzero((char *) &server, sizeof(server));
  server.sin_family = AF_INET;
  doit( argc, argv) ;
  fprintf( stderr, "mdate: all hosts failed\n");
  exit(1); 
}

/*
 * All these functions will exit() on success and return only on failure
 */
void 
doit( argc, argv)
     int argc;
     char **argv ;
{
  struct hostent *hp;
  char **ap;

  for (; optind < argc; optind++) 
    {
      /* See if an explicit addr was given. If not, try a hostname. */
      server.sin_addr.s_addr = inet_addr(argv[optind]);
      if (server.sin_addr.s_addr != (unsigned long)-1) 
	{
	  hostname = argv[optind] ;
	  if ( dflag>1) printf( "mdate: Trying %s... \n", hostname );
	  try_host() ;
	} 
      else 
	{
	  hp = gethostbyname(argv[optind]);
	  if (hp == NULL) 
	    {
	      fprintf(stderr, "mdate: host %s is unknown.\n", argv[optind]);
	      continue;
	    }

	  /*
	   * There are 2 versions of the struct returned by gethostbyname:
	   * one has a single host address; the other has a list.  
	   */
#ifdef OLDGETHOST
	  ap = (char**) &(hp->h_addr_list); 
#else
	  for (ap = hp->h_addr_list; *ap; ap++) 
#endif
	    {
	      hostname = hp->h_name ;
	      bcopy((char *)*ap, (char *)&server.sin_addr, hp->h_length);
	      if ( dflag>1) printf( "mdate: Trying %s (%s)... \n",
				   hostname, inet_ntoa(server.sin_addr));
	      try_host() ;
	    }
	}
    }
}

void try_host()
{
  int n ;

  for( n=0; n<count; n++)
    {

      if ( (sockfd = socket(AF_INET, SOCK_RAW, proto->p_proto)) < 0)
	{ perror( "mdate: socket"); return ;}

      if( send_packet( n) )
	rcv_packet( n);
    }
  if ( min_trip < MAXINT)
    set_time( ) ;		/* else return */
}

/*
 * Compose and transmit an ICMP TIME REQUEST packet.  The IP header
 * will be prepended by the kernel.  The ID field is our UNIX process ID,
 * and the sequence number is an ascending integer.  Timestamp = milli-
 * seconds since midnight in network byte order.
 */
int
send_packet( seq)
     int seq;			
{
  int i ;
  struct icmp	*icp;		
  struct timeval tv_now;
  long stamp, *n;
  char packet[ PACKSIZE];

  bzero( packet, PACKSIZE);
  icp = (struct icmp *) packet; /* pointer to ICMP header */
  icp->icmp_type  = ICMP_TSTAMP;
  icp->icmp_code  = 0;
  icp->icmp_cksum = 0;		/* Provisional: checksum filled in later */
  icp->icmp_id    = ident;	/* our pid, to identify on return */
  icp->icmp_seq   = seq;	/* sequence number */

  if ( gettimeofday( &tv_now, (struct timezone*) 0 ) <0)
    { perror( "mdate: gettimeofday_2"); return(0);}

  stamp = htonl( millis(tv_now)  ) ;
  icp->icmp_otime = (n_time) stamp ;
  icp->icmp_cksum = in_cksum( (u_short*) icp, PACKSIZE);

  i = sendto(sockfd, packet, PACKSIZE, 0,
	     (struct sockaddr *) &server, sizeof( server));

  if (i < 0) { perror("mdate: sendto"); return(0);}
  else if ( i != PACKSIZE || dflag > 2)
  {
    printf(  "\n^mdate: sent %d bytes to %s  at %s ", 
	    i, hostname, ctime( (time_t *) &tv_now.tv_sec) );
    n= (long*) packet;
    for ( i=0; i<8; i++)
      printf( "  %8.8x", *(n+i) );
    printf( "\n" );
  }
  return(1);
}

/*
 * Checksum routine for Internet Protocol family headers (C Version).
 *
 * Refer to "Computing the Internet Checksum" by R. Braden, D. Borman and
 * C. Partridge, Computer Communication Review, Vol. 19, No. 2, April 1989,
 * pp. 86-101.
 */
int
in_cksum(ptr, nbytes)
     register u_short *ptr;
     register int nbytes;
{
  register long sum;		/* assumes long == 32 bits */
  u_short oddbyte;
  register u_short answer;	/* assumes u_short == 16 bits */

  /*
   * Our algorithm is simple, using a 32-bit accumulator (sum),
   * we add sequential 16-bit words to it, and at the end, fold back
   * all the carry bits from the top 16 bits into the lower 16 bits.
   */

  sum = 0;
  while (nbytes > 1)  
    {
      sum += *ptr++;
      nbytes -= 2;
    }

  /* mop up an odd byte, if necessary */
  if (nbytes == 1) 
    {
      oddbyte = 0;		/* make sure top half is zero */
      *((u_char *) &oddbyte) = *(u_char *)ptr; /* one byte only */
      sum += oddbyte;
    }

  /* Add back carry outs from top 16 bits to low 16 bits. */

  sum  = (sum >> 16) + (sum & SHORT_MASK ); /* add high-16 to low-16 */
  sum += (sum >> 16);		/* add carry */
  answer = ~sum;		/* ones-complement, then truncate to 16 bits */
  return(answer);
}

/*
 * Receive ICMP replies.
 */
int
rcv_packet( seq)
     int seq;			
{
  char		recvpack[MAXPACKET];	
  register int		n;
  int			fromlen;
  struct sockaddr_in	from;
  fd_set readfds;
  struct icmp *ptr;
  struct timeval tout,  etv ;
  int triptime, diff, otime, rtime, ttime, etime;

  tout.tv_sec = READTOUT;
  tout.tv_usec = (long)0;
  FD_ZERO(&readfds);
  FD_SET(sockfd, &readfds);

  while (1)
    {
/*    n=select( sockfd+1, (int *)&readfds, (fd_set*) 0, (fd_set*) 0, &tout); */
      n=select( sockfd+1, (int *)&readfds, (int *) 0, (int *) 0, &tout);
  
    if (n<0) { perror( "mdate: select"); return(0); }
      else if ( n==0) break;
      fromlen = sizeof(from);
      if ( (n = recvfrom(sockfd, recvpack, sizeof(recvpack), 0,
			 (struct sockaddr *) &from, &fromlen)) < 0)
	{ perror( "mdate: recvfrom"); return(0);}
      if ( gettimeofday( &etv, (struct timezone*) 0 ) <0)
	{ perror( "mdate: gettimeofday_3"); return(0);}
      etime = millis( etv) ;
      if (dflag>2)
	printf( "mdate: received %d bytes (etime=0x%x) at %s",
	       n, etime, ctime( (time_t *) &etv.tv_sec) );

      if ( ( ptr = pr_pack(recvpack, n, seq, &from))
	  !=  (struct icmp*) NULL )
	{

	  otime = ntohl( ptr->icmp_otime) ;
	  rtime = ntohl( ptr->icmp_rtime) ;
	  ttime = ntohl( ptr->icmp_ttime) ;

	  triptime = max( etime - otime , (int) QUANTUM/2) ;
	  diff = ( rtime + ttime - etime - otime) /2 ;
	  if( dflag>1)
	    {
	      if(dflag>2 || 0 == seq)
		printf( "   local time       remote time       trip time   diff   (m-sec)\n");
	      printf( "%12d      %12d %12d %12d\n", otime, rtime, triptime, diff) ;
	      replies ++ ; sum_trip+=triptime ;
	    }
	  if ( triptime < min_trip)
	    {
	      min_trip = triptime ;
	      best_diff = diff ;
	    }
	  return(1);
	}
    }
  return(0);
}

/*
 * Decide if a packet is the one we want.  This logic is necessary
 * because ALL readers of the ICMP socket get a copy of ALL ICMP packets
 * which arrive ('tis only fair).  This permits multiple copies of this
 * program to be run without having intermingled output (or statistics!).
 */
struct icmp *
pr_pack(buf, cc, seq, from )
     char *buf;			/* ptr to start of IP header */
     int cc,			/* total size of received packet */
       seq ;			/* Sequence no. */
     struct sockaddr_in	*from;	/* address of sender */
{
  int iphdrlen ;
  struct ip *iphd;		/* ptr to IP header */
  register struct icmp	*icp;	/* ptr to ICMP header */
  long *lp, n ;
  struct timeval tv;

  if ( gettimeofday( &tv, (struct timezone*) 0 ) <0)
    { perror( "mdate: gettimeofday_4"); return(0);}

  /*
   * We have to look at the IP header, to get its length.
   * We also verify that what follows the IP header contains at
   * least an ICMP header (8 bytes minimum). Unfortunately
   * some systems DO and others DONT strip the IP header. 
   */
  iphd = (struct ip *) buf;

  if( ( iphd->ip_v == IPVERSION ) && ( iphd->ip_p == proto->p_proto) )
    iphdrlen = iphd->ip_hl << 2;	/* convert # 32-bit words to #bytes */
  else
    iphdrlen=0;

  if (cc < iphdrlen + ICMP_MINLEN)
    {
      fprintf(stderr, "packet too short (%d bytes) from %s\n", cc,
	      inet_ntoa( from->sin_addr ) );
      return( NULL);
    }
  cc -= iphdrlen;
  icp = (struct icmp *)(buf + iphdrlen);

  if ( dflag > 2)
    /* Print first words of ICMP pkt */
    {
      lp = (long *) buf;	
      if( iphdrlen >0)
	{
	  for ( n=0; n<(iphdrlen>>2) ; n++, lp++)
	    printf( "  %8.8x", *lp );
	  printf( "\n\n" );
	}
      for ( n=0; n<5 ; n++, lp++)
	printf( "  %8.8x", *lp );
      printf( "\n" );
    }

  if ( (icp->icmp_type != ICMP_TSTAMPREPLY)
      || (icp->icmp_id != ident) 
      || (icp->icmp_seq != seq) )
    return ( NULL) ;
  else return( icp);
}

/*
 * Convert an ICMP "type" field to a printable string.
 */
char *
pr_type(t)
register int t;
{
  static char	*ttab[] = 
    {
      "Echo Reply",
      "ICMP 1",
      "ICMP 2",
      "Dest Unreachable",
      "Source Quench",
      "Redirect",
      "ICMP 6",
      "ICMP 7",
      "Echo",
      "ICMP 9",
      "ICMP 10",
      "Time Exceeded",
      "Parameter Problem",
      "Timestamp",
      "Timestamp Reply",
      "Info Request",
      "Info Reply",
      "Mask request",
      "Mask Reply"
      };

  if (t < 0 || t > 18 )
    return("UNKNOWN TYPE");

  return(ttab[t]);
}

/* 
 * Set local time (perhaps).
*/
void
set_time( ) 
{
  struct timeval old, new;
  long ds, dm, s, m;
  char localhost[ MAXHOSTNAMELEN];

  if(  gethostname( localhost, MAXHOSTNAMELEN) <0)
    perror( "mdate: gethostname" );
  min_trip = max( QUANTUM, min_trip) ;
  if ( gettimeofday( &old, (struct timezone*) 0 ) <0)
    { perror( "mdate: gettimeofday_5"); return;}

  if ( abs( best_diff) > HUGE)
    {
      fprintf( stderr, "mdate: error: host %s: huge difference (ignored)\n",
	      hostname);
      exit( 0) ;
    }
  if ( abs( best_diff) < tolerance * min_trip  || ( !aflag && !sflag))
    {
      if (dflag) printf( "mdate: local %s, server %s: diff %d msec: trip %d: time unchanged at %s\n",
			localhost, hostname, best_diff, min_trip, ctime((time_t *) &(old.tv_sec) ) );
      if( dflag>1) printf( "Mean trip time: %d msec\n",
			  (int) sum_trip/replies );
      exit( 0) ;
    }

  ds = (long)( best_diff / K) ;
  dm = (long)( ( best_diff % K) * K) ;
  s = old.tv_sec + ds ;
  m = old.tv_usec + dm ;

#ifndef NOADJTIME
  if ( aflag)
    {
      new.tv_sec = ds ;
      new.tv_usec =  dm ;
      if ( adjtime( &new, (struct timeval *) NULL) < 0)
	{ perror( "mdate: adjtime"); return;}
    }
  else 
#endif
    if ( sflag)
      {
	if ( m<0)
	  { m+=M ; s--; }
	if ( m>= M )
	  { m-=M ; s++; }
	new.tv_sec = s ;
	new.tv_usec = m ;
	if ( settimeofday( &new, (struct timezone *) NULL) < 0)
	  { perror( "mdate: settimeofday"); return;}
      }
    else fprintf( stderr, "mdate: internal bug report\n" );  
  if (dflag) printf( "mdate: local %s, server %s: diff %d msec: trip %d: new time %s\n",
		    localhost, hostname, best_diff, min_trip, ctime( &s) );
  exit( 0) ;
}

#ifdef NORANDOM

/* 16-bit pseudo-random numbers, needed for HP's Instant UG.
   Loosely based on BSD rand(3), but that had some uglies. 
*/

#define P 65537
#define T 43063
#define Q 37813

int next ;

void srandom( seed)
int seed;
{
  next = (  ( seed / P ) * Q  + ( seed % P ) * T  ) % P ;
  random() ; random() ;
}

int random( void)
{
 next = ( next * T + Q ) % P ;
 return( next);
}

#endif
